@@ -3,11 +3,11 @@
  *
  * Copyright (C) 1999 Weiss
  * Copyright (C) 2004 Chris Shoemaker
- * Copyright (C) 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2004-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -16,8 +16,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
@@ -32,15 +31,25 @@ extern int preserve_hard_links;
 extern int preserve_devices;
 extern int preserve_uid;
 extern int preserve_gid;
+extern int preserve_acls;
+extern int preserve_xattrs;
 extern int always_checksum;
 extern int do_compression;
-extern int def_compress_level;
+extern int inplace;
+extern int append_mode;
 extern int protocol_version;
 extern char *batch_name;
+#ifdef ICONV_OPTION
+extern char *iconv_opt;
+#endif
 
 extern struct filter_list_struct filter_list;
 
-static int tweaked_compress_level;
+int batch_stream_flags;
+
+static int tweaked_append;
+static int tweaked_append_verify;
+static int tweaked_iconv;
 
 static int *flag_ptr[] = {
 	&recurse,		/* 0 */
@@ -51,7 +60,13 @@ static int *flag_ptr[] = {
 	&preserve_hard_links,	/* 5 */
 	&always_checksum,	/* 6 */
 	&xfer_dirs,		/* 7 (protocol 29) */
-	&tweaked_compress_level,/* 8 (protocol 29) */
+	&do_compression,	/* 8 (protocol 29) */
+	&tweaked_iconv,		/* 9  (protocol 30) */
+	&preserve_acls,		/* 10 (protocol 30) */
+	&preserve_xattrs,	/* 11 (protocol 30) */
+	&inplace,		/* 12 (protocol 30) */
+	&tweaked_append,	/* 13 (protocol 30) */
+	&tweaked_append_verify,	/* 14 (protocol 30) */
 	NULL
 };
 
@@ -65,6 +80,12 @@ static char *flag_name[] = {
 	"--checksum (-c)",
 	"--dirs (-d)",
 	"--compress (-z)",
+	"--iconv",
+	"--acls (-A)",
+	"--xattrs (-X)",
+	"--inplace",
+	"--append",
+	"--append-verify",
 	NULL
 };
 
@@ -72,16 +93,14 @@ void write_stream_flags(int fd)
 {
 	int i, flags;
 
-#if Z_DEFAULT_COMPRESSION == -1
-	tweaked_compress_level = do_compression ? def_compress_level + 2 : 0;
-#else
-#error internal logic error!  Fix def_compress_level logic above and below too!
+	tweaked_append = append_mode == 1;
+	tweaked_append_verify = append_mode == 2;
+#ifdef ICONV_OPTION
+	tweaked_iconv = iconv_opt != NULL;
 #endif
 
 	/* Start the batch file with a bitmap of data-stream-affecting
 	 * flags. */
-	if (protocol_version < 29)
-		flag_ptr[7] = NULL;
 	for (i = 0, flags = 0; flag_ptr[i]; i++) {
 		if (*flag_ptr[i])
 			flags |= 1 << i;
@@ -91,13 +110,31 @@ void write_stream_flags(int fd)
 
 void read_stream_flags(int fd)
 {
-	int i, flags;
+	batch_stream_flags = read_int(fd);
+}
+
+void check_batch_flags(void)
+{
+	int i;
 
 	if (protocol_version < 29)
 		flag_ptr[7] = NULL;
-	for (i = 0, flags = read_int(fd); flag_ptr[i]; i++) {
-		int set = flags & (1 << i) ? 1 : 0;
+	else if (protocol_version < 30)
+		flag_ptr[9] = NULL;
+	tweaked_append = append_mode == 1;
+	tweaked_append_verify = append_mode == 2;
+#ifdef ICONV_OPTION
+	tweaked_iconv = iconv_opt != NULL;
+#endif
+	for (i = 0; flag_ptr[i]; i++) {
+		int set = batch_stream_flags & (1 << i) ? 1 : 0;
 		if (*flag_ptr[i] != set) {
+			if (i == 9) {
+				rprintf(FERROR,
+					"%s specify the --iconv option to use this batch file.\n",
+					set ? "Please" : "Do not");
+				exit_cleanup(RERR_SYNTAX);
+			}
 			if (verbose) {
 				rprintf(FINFO,
 					"%sing the %s option to match the batchfile.\n",
@@ -113,12 +150,10 @@ void read_stream_flags(int fd)
 			xfer_dirs = 0;
 	}
 
-	if (tweaked_compress_level == 0 || tweaked_compress_level == 2)
-		do_compression = 0;
-	else {
-		do_compression = 1;
-		def_compress_level = tweaked_compress_level - 2;
-	}
+	if (tweaked_append)
+		append_mode = 1;
+	else if (tweaked_append_verify)
+		append_mode = 2;
 }
 
 static void write_arg(int fd, char *arg)
@@ -180,7 +215,7 @@ void write_batch_shell_file(int argc, char *argv[], int file_arg_cnt)
 	if (fd < 0) {
 		rsyserr(FERROR, errno, "Batch file %s open error",
 			filename);
-		exit_cleanup(1);
+		exit_cleanup(RERR_FILESELECT);
 	}
 
 	/* Write argvs info to BATCH.sh file */
@@ -226,6 +261,6 @@ void write_batch_shell_file(int argc, char *argv[], int file_arg_cnt)
 	if (write(fd, "\n", 1) != 1 || close(fd) < 0) {
 		rsyserr(FERROR, errno, "Batch file %s write error",
 			filename);
-		exit_cleanup(1);
+		exit_cleanup(RERR_FILEIO);
 	}
 }
