@@ -3,11 +3,11 @@
  *
  * Copyright (C) 1998 Andrew Tridgell
  * Copyright (C) 2002 Martin Pool
- * Copyright (C) 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2004-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -16,8 +16,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
@@ -29,47 +28,55 @@
 extern int sparse_files;
 
 static char last_byte;
-static int last_sparse;
+static size_t sparse_seek = 0;
 
 int sparse_end(int f)
 {
-	if (last_sparse) {
-		do_lseek(f,-1,SEEK_CUR);
-		return (write(f,&last_byte,1) == 1 ? 0 : -1);
-	}
-	last_sparse = 0;
-	return 0;
+	int ret;
+
+	if (!sparse_seek)
+		return 0;
+
+	do_lseek(f, sparse_seek-1, SEEK_CUR);
+	sparse_seek = 0;
+
+	do {
+		ret = write(f, "", 1);
+	} while (ret < 0 && errno == EINTR);
+
+	return ret <= 0 ? -1 : 0;
 }
 
 
-static int write_sparse(int f,char *buf,size_t len)
+static int write_sparse(int f, char *buf, size_t len)
 {
-	size_t l1=0, l2=0;
+	size_t l1 = 0, l2 = 0;
 	int ret;
 
 	for (l1 = 0; l1 < len && buf[l1] == 0; l1++) {}
 	for (l2 = 0; l2 < len-l1 && buf[len-(l2+1)] == 0; l2++) {}
 
+	/* XXX Riddle me this: why does this function SLOW DOWN when I
+	 * remove the following (unneeded) line?? Core Duo weirdness? */
 	last_byte = buf[len-1];
 
-	if (l1 == len || l2 > 0)
-		last_sparse=1;
-
-	if (l1 > 0) {
-		do_lseek(f,l1,SEEK_CUR);
-	}
+	sparse_seek += l1;
 
 	if (l1 == len)
 		return len;
 
-	ret = write(f, buf + l1, len - (l1+l2));
-	if (ret == -1 || ret == 0)
-		return ret;
-	else if (ret != (int) (len - (l1+l2)))
-		return (l1+ret);
+	if (sparse_seek)
+		do_lseek(f, sparse_seek, SEEK_CUR);
+	sparse_seek = l2;
 
-	if (l2 > 0)
-		do_lseek(f,l2,SEEK_CUR);
+	while ((ret = write(f, buf + l1, len - (l1+l2))) <= 0) {
+		if (ret < 0 && errno == EINTR)
+			continue;
+		return ret;
+	}
+
+	if (ret != (int)(len - (l1+l2)))
+		return l1+ret;
 
 	return len;
 }
@@ -107,7 +114,7 @@ int write_file(int f,char *buf,size_t len)
 
 	while (len > 0) {
 		int r1;
-		if (sparse_files) {
+		if (sparse_files > 0) {
 			int len1 = MIN(len, SPARSE_WRITE_SIZE);
 			r1 = write_sparse(f, buf, len1);
 		} else {
@@ -152,13 +159,12 @@ struct map_struct *map_file(int fd, OFF_T len, int32 read_size,
 {
 	struct map_struct *map;
 
-	if (!(map = new(struct map_struct)))
+	if (!(map = new0(struct map_struct)))
 		out_of_memory("map_file");
 
 	if (blk_size && (read_size % blk_size))
 		read_size += blk_size - (read_size % blk_size);
 
-	memset(map, 0, sizeof map[0]);
 	map->fd = fd;
 	map->file_size = len;
 	map->def_window_size = read_size;
@@ -190,7 +196,7 @@ char *map_ptr(struct map_struct *map, OFF_T offset, int32 len)
 	window_start = offset;
 	window_size = map->def_window_size;
 	if (window_start + window_size > map->file_size)
-		window_size = map->file_size - window_start;
+		window_size = (int32)(map->file_size - window_start);
 	if (len > window_size)
 		window_size = len;
 
@@ -208,7 +214,7 @@ char *map_ptr(struct map_struct *map, OFF_T offset, int32 len)
 	    window_start < map->p_offset + map->p_len &&
 	    window_start + window_size >= map->p_offset + map->p_len) {
 		read_start = map->p_offset + map->p_len;
-		read_offset = read_start - window_start;
+		read_offset = (int32)(read_start - window_start);
 		read_size = window_size - read_offset;
 		memmove(map->p, map->p + (map->p_len - read_offset), read_offset);
 	} else {
