@@ -697,8 +697,11 @@ void itemize(const char *fnamecmp, struct file_struct *file, int ndx, int statre
 			if (iflags & ITEM_XNAME_FOLLOWS)
 				write_vstring(sock_f_out, xname, strlen(xname));
 #ifdef SUPPORT_XATTRS
-			if (iflags & ITEM_REPORT_XATTR && !dry_run)
-				send_xattr_request(NULL, file, sock_f_out);
+			if (preserve_xattrs && !dry_run
+			 && iflags & (ITEM_REPORT_XATTR|ITEM_TRANSFER)) {
+				send_xattr_request(NULL, file,
+					iflags & ITEM_REPORT_XATTR ? sock_f_out : -1);
+			}
 #endif
 		} else if (ndx >= 0) {
 			enum logcode code = logfile_format_has_i ? FINFO : FCLIENT;
@@ -753,18 +756,19 @@ static void sum_sizes_sqroot(struct sum_struct *sum, int64 len)
 {
 	int32 blength;
 	int s2length;
+	int64 l;
 
 	if (block_size)
 		blength = block_size;
 	else if (len <= BLOCK_SIZE * BLOCK_SIZE)
 		blength = BLOCK_SIZE;
 	else {
+		int32 max_blength = protocol_version < 30 ? OLD_MAX_BLOCK_SIZE : MAX_BLOCK_SIZE;
 		int32 c;
-		int64 l;
 		int cnt;
 		for (c = 1, l = len, cnt = 0; l >>= 2; c <<= 1, cnt++) {}
-		if (cnt >= 31 || c >= MAX_BLOCK_SIZE)
-			blength = MAX_BLOCK_SIZE;
+		if (c < 0 || c >= max_blength)
+			blength = max_blength;
 		else {
 		    blength = 0;
 		    do {
@@ -783,7 +787,6 @@ static void sum_sizes_sqroot(struct sum_struct *sum, int64 len)
 		s2length = SUM_LENGTH;
 	} else {
 		int32 c;
-		int64 l;
 		int b = BLOCKSUM_BIAS;
 		for (l = len; l >>= 1; b += 2) {}
 		for (c = blength; (c >>= 1) && b; b--) {}
@@ -797,7 +800,10 @@ static void sum_sizes_sqroot(struct sum_struct *sum, int64 len)
 	sum->blength	= blength;
 	sum->s2length	= s2length;
 	sum->remainder	= (int32)(len % blength);
-	sum->count	= (int32)(len / blength) + (sum->remainder != 0);
+	sum->count	= (int32)(l = (len / blength) + (sum->remainder != 0));
+
+	if ((int64)sum->count != l)
+		sum->count = -1;
 
 	if (sum->count && verbose > 2) {
 		rprintf(FINFO,
@@ -813,7 +819,7 @@ static void sum_sizes_sqroot(struct sum_struct *sum, int64 len)
  *
  * Generate approximately one checksum every block_len bytes.
  */
-static void generate_and_send_sums(int fd, OFF_T len, int f_out, int f_copy)
+static int generate_and_send_sums(int fd, OFF_T len, int f_out, int f_copy)
 {
 	int32 i;
 	struct map_struct *mapbuf;
@@ -821,10 +827,12 @@ static void generate_and_send_sums(int fd, OFF_T len, int f_out, int f_copy)
 	OFF_T offset = 0;
 
 	sum_sizes_sqroot(&sum, len);
+	if (sum.count < 0)
+		return -1;
 	write_sum_head(f_out, &sum);
 
 	if (append_mode > 0 && f_copy < 0)
-		return;
+		return 0;
 
 	if (len > 0)
 		mapbuf = map_file(fd, len, MAX_MAP_SIZE, sum.blength);
@@ -861,6 +869,8 @@ static void generate_and_send_sums(int fd, OFF_T len, int f_out, int f_copy)
 
 	if (mapbuf)
 		unmap_file(mapbuf);
+
+	return 0;
 }
 
 
@@ -922,6 +932,7 @@ static int copy_altdest_file(const char *src, const char *dest, struct file_stru
 {
 	char buf[MAXPATHLEN];
 	const char *copy_to, *partialptr;
+	int save_preserve_xattrs = preserve_xattrs;
 	int ok, fd_w;
 
 	if (inplace) {
@@ -946,7 +957,9 @@ static int copy_altdest_file(const char *src, const char *dest, struct file_stru
 		return -1;
 	}
 	partialptr = partial_dir ? partial_dir_fname(dest) : NULL;
+	preserve_xattrs = 0; /* xattrs were copied with file */
 	ok = finish_transfer(dest, copy_to, src, partialptr, file, 1, 0);
+	preserve_xattrs = save_preserve_xattrs;
 	cleanup_disable();
 	return ok ? 0 : -1;
 }
@@ -1284,7 +1297,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		skip_dir = NULL;
 	}
 
-	if (daemon_filter_list.head) {
+	if (daemon_filter_list.head && (*fname != '.' || fname[1])) {
 		if (check_filter(&daemon_filter_list, FLOG, fname, is_dir) < 0) {
 			if (is_dir < 0)
 				return;
@@ -1432,6 +1445,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 			if (j == -2) {
 				itemizing = 0;
 				code = FNONE;
+				statret = 1;
 			} else if (j >= 0)
 				statret = 1;
 		}
@@ -1454,6 +1468,10 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 				goto cleanup;
 			}
 		}
+#ifdef SUPPORT_XATTRS
+		if (preserve_xattrs && statret == 1)
+			copy_xattrs(fnamecmpbuf, fname);
+#endif
 		if (set_file_attrs(fname, file, real_ret ? NULL : &real_sx, NULL, 0)
 		    && verbose && code != FNONE && f_out != -1)
 			rprintf(code, "%s/\n", fname);
@@ -1868,15 +1886,21 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 			close(fd);
 			goto cleanup;
 		}
-		if ((f_copy = do_open(backupptr, O_WRONLY | O_CREAT | O_TRUNC | O_EXCL, 0600)) < 0
-		 && (errno != ENOENT || make_bak_dir(backupptr) < 0
-		  || (f_copy = do_open(backupptr, O_WRONLY | O_CREAT | O_TRUNC | O_EXCL, 0600)) < 0)) {
-			rsyserr(FERROR_XFER, errno, "open %s",
-				full_fname(backupptr));
-			unmake_file(back_file);
-			back_file = NULL;
-			close(fd);
-			goto cleanup;
+		if ((f_copy = do_open(backupptr, O_WRONLY | O_CREAT | O_TRUNC | O_EXCL, 0600)) < 0) {
+			int save_errno = errno ? errno : EINVAL; /* 0 paranoia */
+			if (errno == ENOENT && make_bak_dir(backupptr) == 0) {
+				if ((f_copy = do_open(backupptr, O_WRONLY | O_CREAT | O_TRUNC | O_EXCL, 0600)) < 0)
+					save_errno = errno ? errno : save_errno;
+				else
+					save_errno = 0;
+			}
+			if (save_errno) {
+				rsyserr(FERROR_XFER, save_errno, "open %s", full_fname(backupptr));
+				unmake_file(back_file);
+				back_file = NULL;
+				close(fd);
+				goto cleanup;
+			}
 		}
 		fnamecmp_type = FNAMECMP_BACKUP;
 	}
@@ -1931,16 +1955,32 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 
 	if (statret != 0 || whole_file)
 		write_sum_head(f_out, NULL);
-	else {
-		generate_and_send_sums(fd, sx.st.st_size, f_out, f_copy);
+	else if (sx.st.st_size <= 0) {
+		write_sum_head(f_out, NULL);
+		close(fd);
+	} else {
+		if (generate_and_send_sums(fd, sx.st.st_size, f_out, f_copy) < 0) {
+			rprintf(FWARNING,
+			    "WARNING: file is too large for checksum sending: %s\n",
+			    fnamecmp);
+			write_sum_head(f_out, NULL);
+		}
 		close(fd);
 	}
 
   cleanup:
 	if (back_file) {
+		int save_preserve_xattrs = preserve_xattrs;
 		if (f_copy >= 0)
 			close(f_copy);
+#ifdef SUPPORT_XATTRS
+		if (preserve_xattrs) {
+			copy_xattrs(fname, backupptr);
+			preserve_xattrs = 0;
+		}
+#endif
 		set_file_attrs(backupptr, back_file, NULL, NULL, 0);
+		preserve_xattrs = save_preserve_xattrs;
 		if (verbose > 1) {
 			rprintf(FINFO, "backed up %s to %s\n",
 				fname, backupptr);
@@ -2179,7 +2219,10 @@ void generate_files(int f_out, const char *local_name)
 
 		if (inc_recurse && cur_flist->parent_ndx >= 0) {
 			struct file_struct *fp = dir_flist->files[cur_flist->parent_ndx];
-			f_name(fp, fbuf);
+			if (solo_file)
+				strlcpy(fbuf, solo_file, sizeof fbuf);
+			else
+				f_name(fp, fbuf);
 			ndx = cur_flist->ndx_start - 1;
 			recv_generator(fbuf, fp, ndx, itemizing, code, f_out);
 			if (delete_during && dry_run < 2 && !list_only
