@@ -561,8 +561,12 @@ int finish_transfer(const char *fname, const char *fnametmp,
 		goto do_set_file_attrs;
 	}
 
-	if (make_backups > 0 && overwriting_basis && !make_backup(fname))
-		return 1;
+	if (make_backups > 0 && overwriting_basis) {
+		if (!make_backup(fname))
+			return 1;
+		if (fnamecmp == fname)
+			fnamecmp = get_backup_name(fname);
+	}
 
 	/* Change permissions before putting the file into place. */
 	set_file_attrs(fnametmp, file, NULL, fnamecmp,
