@@ -3,11 +3,11 @@
  *
  * Copyright (C) 1996 Andrew Tridgell
  * Copyright (C) 1996 Paul Mackerras
- * Copyright (C) 2003, 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2003-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -16,16 +16,17 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
 
 extern int verbose;
+extern int dry_run;
 extern int do_xfers;
 extern int am_server;
 extern int am_daemon;
+extern int inc_recurse;
 extern int log_before_transfer;
 extern int stdout_format_has_i;
 extern int logfile_format_has_i;
@@ -33,6 +34,7 @@ extern int csum_length;
 extern int append_mode;
 extern int io_error;
 extern int allowed_lull;
+extern int preserve_xattrs;
 extern int protocol_version;
 extern int remove_source_files;
 extern int updating_basis_file;
@@ -42,9 +44,7 @@ extern int inplace;
 extern int batch_fd;
 extern int write_batch;
 extern struct stats stats;
-extern struct file_list *the_file_list;
-extern char *stdout_format;
-
+extern struct file_list *cur_flist, *first_flist, *dir_flist;
 
 /**
  * @file
@@ -76,7 +76,7 @@ static struct sum_struct *receive_sums(int f)
 			(double)s->count, (long)s->blength, (long)s->remainder);
 	}
 
-	if (append_mode) {
+	if (append_mode > 0) {
 		s->flength = (OFF_T)s->count * s->blength;
 		if (s->remainder)
 			s->flength -= s->blength - s->remainder;
@@ -122,31 +122,35 @@ void successful_send(int ndx)
 {
 	char fname[MAXPATHLEN];
 	struct file_struct *file;
-	unsigned int offset;
+	struct file_list *flist;
 
-	if (ndx < 0 || ndx >= the_file_list->count)
+	if (!remove_source_files)
 		return;
 
-	file = the_file_list->files[ndx];
-	if (file->dir.root) {
-		offset = stringjoin(fname, sizeof fname,
-				    file->dir.root, "/", NULL);
-	} else
-		offset = 0;
-	f_name(file, fname + offset);
-	if (remove_source_files) {
-		if (do_unlink(fname) == 0) {
-			if (verbose > 1)
-				rprintf(FINFO, "sender removed %s\n", fname + offset);
-		} else
-			rsyserr(FERROR, errno, "sender failed to remove %s", fname + offset);
+	if (!(flist = flist_for_ndx(ndx))) {
+		rprintf(FERROR,
+			"INTERNAL ERROR: unable to find flist for item %d\n",
+			ndx);
+		return;
 	}
+
+	file = flist->files[ndx - flist->ndx_start];
+	if (!change_pathname(file, NULL, 0))
+		return;
+	f_name(file, fname);
+
+	if (do_unlink(fname) == 0) {
+		if (verbose > 1)
+			rprintf(FINFO, "sender removed %s\n", fname);
+	} else
+		rsyserr(FERROR, errno, "sender failed to remove %s", fname);
 }
 
 static void write_ndx_and_attrs(int f_out, int ndx, int iflags,
+				const char *fname, struct file_struct *file,
 				uchar fnamecmp_type, char *buf, int len)
 {
-	write_int(f_out, ndx);
+	write_ndx(f_out, ndx);
 	if (protocol_version < 29)
 		return;
 	write_shortint(f_out, iflags);
@@ -154,118 +158,85 @@ static void write_ndx_and_attrs(int f_out, int ndx, int iflags,
 		write_byte(f_out, fnamecmp_type);
 	if (iflags & ITEM_XNAME_FOLLOWS)
 		write_vstring(f_out, buf, len);
+#ifdef SUPPORT_XATTRS
+	if (preserve_xattrs && iflags & ITEM_REPORT_XATTR && !dry_run)
+		send_xattr_request(fname, file, f_out);
+#endif
 }
 
-/* This is also used by receive.c with f_out = -1. */
-int read_item_attrs(int f_in, int f_out, int ndx, uchar *type_ptr,
-		    char *buf, int *len_ptr)
-{
-	int len;
-	uchar fnamecmp_type = FNAMECMP_FNAME;
-	int iflags = protocol_version >= 29 ? read_shortint(f_in)
-		   : ITEM_TRANSFER | ITEM_MISSING_DATA;
-
-	/* Handle the new keep-alive (no-op) packet. */
-	if (ndx == the_file_list->count && iflags == ITEM_IS_NEW)
-		;
-	else if (ndx < 0 || ndx >= the_file_list->count) {
-		rprintf(FERROR, "Invalid file index: %d (count=%d) [%s]\n",
-			ndx, the_file_list->count, who_am_i());
-		exit_cleanup(RERR_PROTOCOL);
-	} else if (iflags == ITEM_IS_NEW) {
-		rprintf(FERROR, "Invalid itemized flag word: %x [%s]\n",
-			iflags, who_am_i());
-		exit_cleanup(RERR_PROTOCOL);
-	}
-
-	if (iflags & ITEM_BASIS_TYPE_FOLLOWS)
-		fnamecmp_type = read_byte(f_in);
-	*type_ptr = fnamecmp_type;
-
-	if (iflags & ITEM_XNAME_FOLLOWS) {
-		if ((len = read_vstring(f_in, buf, MAXPATHLEN)) < 0)
-			exit_cleanup(RERR_PROTOCOL);
-	} else {
-		*buf = '\0';
-		len = -1;
-	}
-	*len_ptr = len;
-
-	if (iflags & ITEM_TRANSFER) {
-		if (!S_ISREG(the_file_list->files[ndx]->mode)) {
-			rprintf(FERROR,
-				"received request to transfer non-regular file: %d [%s]\n",
-				ndx, who_am_i());
-			exit_cleanup(RERR_PROTOCOL);
-		}
-	} else if (f_out >= 0) {
-		write_ndx_and_attrs(f_out, ndx, iflags,
-				    fnamecmp_type, buf, len);
-	}
-
-	return iflags;
-}
-
-void send_files(struct file_list *flist, int f_out, int f_in)
+void send_files(int f_in, int f_out)
 {
 	int fd = -1;
 	struct sum_struct *s;
 	struct map_struct *mbuf = NULL;
 	STRUCT_STAT st;
-	char *fname2, fname[MAXPATHLEN];
-	char xname[MAXPATHLEN];
+	char fname[MAXPATHLEN], xname[MAXPATHLEN];
+	const char *path, *slash;
 	uchar fnamecmp_type;
 	int iflags, xlen;
 	struct file_struct *file;
 	int phase = 0, max_phase = protocol_version >= 29 ? 2 : 1;
 	struct stats initial_stats;
-	int save_make_backups = make_backups;
 	int itemizing = am_server ? logfile_format_has_i : stdout_format_has_i;
 	enum logcode log_code = log_before_transfer ? FLOG : FINFO;
 	int f_xfer = write_batch < 0 ? batch_fd : f_out;
-	int i, j;
+	int ndx, j;
 
 	if (verbose > 2)
 		rprintf(FINFO, "send_files starting\n");
 
 	while (1) {
-		unsigned int offset;
+		if (inc_recurse)
+			send_extra_file_list(f_out, FILECNT_LOOKAHEAD);
 
-		i = read_int(f_in);
-		if (i == -1) {
+		/* This call also sets cur_flist. */
+		ndx = read_ndx_and_attrs(f_in, &iflags, &fnamecmp_type,
+					 xname, &xlen);
+		if (ndx == NDX_DONE) {
+			if (inc_recurse && first_flist) {
+				flist_free(first_flist);
+				if (first_flist) {
+					write_ndx(f_out, NDX_DONE);
+					continue;
+				}
+			}
 			if (++phase > max_phase)
 				break;
-			csum_length = SUM_LENGTH;
 			if (verbose > 2)
 				rprintf(FINFO, "send_files phase=%d\n", phase);
-			write_int(f_out, -1);
-			/* For inplace: redo phase turns off the backup
-			 * flag so that we do a regular inplace send. */
-			make_backups = 0;
-			append_mode = 0;
+			write_ndx(f_out, NDX_DONE);
 			continue;
 		}
 
-		iflags = read_item_attrs(f_in, f_out, i, &fnamecmp_type,
-					 xname, &xlen);
-		if (iflags == ITEM_IS_NEW) /* no-op packet */
-			continue;
+		if (inc_recurse)
+			send_extra_file_list(f_out, FILECNT_LOOKAHEAD);
 
-		file = flist->files[i];
-		if (file->dir.root) {
-			/* N.B. We're sure that this fits, so offset is OK. */
-			offset = strlcpy(fname, file->dir.root, sizeof fname);
-			if (!offset || fname[offset-1] != '/')
-				fname[offset++] = '/';
-		} else
-			offset = 0;
-		fname2 = f_name(file, fname + offset);
+		if (ndx - cur_flist->ndx_start >= 0)
+			file = cur_flist->files[ndx - cur_flist->ndx_start];
+		else
+			file = dir_flist->files[cur_flist->parent_ndx];
+		if (F_PATHNAME(file)) {
+			path = F_PATHNAME(file);
+			slash = "/";
+		} else {
+			path = slash = "";
+		}
+		if (!change_pathname(file, NULL, 0))
+			continue;
+		f_name(file, fname);
 
 		if (verbose > 2)
-			rprintf(FINFO, "send_files(%d, %s)\n", i, fname);
+			rprintf(FINFO, "send_files(%d, %s%s%s)\n", ndx, path,slash,fname);
+
+#ifdef SUPPORT_XATTRS
+		if (preserve_xattrs && iflags & ITEM_REPORT_XATTR && !dry_run)
+			recv_xattr_request(file, f_in);
+#endif
 
 		if (!(iflags & ITEM_TRANSFER)) {
 			maybe_log_item(file, iflags, itemizing, xname);
+			write_ndx_and_attrs(f_out, ndx, iflags, fname, file,
+					    fnamecmp_type, xname, xlen);
 			continue;
 		}
 		if (phase == 2) {
@@ -275,17 +246,34 @@ void send_files(struct file_list *flist, int f_out, int f_in)
 			exit_cleanup(RERR_PROTOCOL);
 		}
 
-		updating_basis_file = inplace && (protocol_version >= 29
-			? fnamecmp_type == FNAMECMP_FNAME : !make_backups);
+		if (file->flags & FLAG_FILE_SENT) {
+			if (csum_length == SHORT_SUM_LENGTH) {
+				/* For inplace: redo phase turns off the backup
+				 * flag so that we do a regular inplace send. */
+				make_backups = -make_backups;
+				append_mode = -append_mode;
+				csum_length = SUM_LENGTH;
+			}
+		} else {
+			if (csum_length != SHORT_SUM_LENGTH) {
+				make_backups = -make_backups;
+				append_mode = -append_mode;
+				csum_length = SHORT_SUM_LENGTH;
+			}
+		}
 
-		stats.current_file_index = i;
+		updating_basis_file = inplace && (protocol_version >= 29
+			? fnamecmp_type == FNAMECMP_FNAME : make_backups <= 0);
+
+		if (!am_server && do_progress)
+			set_current_file_index(file, ndx);
 		stats.num_transferred_files++;
-		stats.total_transferred_size += file->length;
+		stats.total_transferred_size += F_LENGTH(file);
 
 		if (!do_xfers) { /* log the transfer */
 			log_item(FCLIENT, file, &stats, iflags, NULL);
-			write_ndx_and_attrs(f_out, i, iflags, fnamecmp_type,
-					    xname, xlen);
+			write_ndx_and_attrs(f_out, ndx, iflags, fname, file,
+					    fnamecmp_type, xname, xlen);
 			continue;
 		}
 
@@ -294,7 +282,7 @@ void send_files(struct file_list *flist, int f_out, int f_in)
 		if (!(s = receive_sums(f_in))) {
 			io_error |= IOERR_GENERAL;
 			rprintf(FERROR, "receive_sums failed\n");
-			return;
+			exit_cleanup(RERR_PROTOCOL);
 		}
 
 		fd = do_open(fname, O_RDONLY, 0);
@@ -302,17 +290,19 @@ void send_files(struct file_list *flist, int f_out, int f_in)
 			if (errno == ENOENT) {
 				enum logcode c = am_daemon
 				    && protocol_version < 28 ? FERROR
-							     : FINFO;
+							     : FWARNING;
 				io_error |= IOERR_VANISHED;
 				rprintf(c, "file has vanished: %s\n",
 					full_fname(fname));
 			} else {
 				io_error |= IOERR_GENERAL;
-				rsyserr(FERROR, errno,
+				rsyserr(FERROR_XFER, errno,
 					"send_files failed to open %s",
 					full_fname(fname));
 			}
 			free_sums(s);
+			if (protocol_version >= 30)
+				send_msg_int(MSG_NO_SEND, ndx);
 			continue;
 		}
 
@@ -322,7 +312,7 @@ void send_files(struct file_list *flist, int f_out, int f_in)
 			rsyserr(FERROR, errno, "fstat failed");
 			free_sums(s);
 			close(fd);
-			return;
+			exit_cleanup(RERR_PROTOCOL);
 		}
 
 		if (st.st_size) {
@@ -332,21 +322,21 @@ void send_files(struct file_list *flist, int f_out, int f_in)
 			mbuf = NULL;
 
 		if (verbose > 2) {
-			rprintf(FINFO, "send_files mapped %s of size %.0f\n",
-				fname, (double)st.st_size);
+			rprintf(FINFO, "send_files mapped %s%s%s of size %.0f\n",
+				path,slash,fname, (double)st.st_size);
 		}
 
-		write_ndx_and_attrs(f_out, i, iflags, fnamecmp_type,
-				    xname, xlen);
+		write_ndx_and_attrs(f_out, ndx, iflags, fname, file,
+				    fnamecmp_type, xname, xlen);
 		write_sum_head(f_xfer, s);
 
 		if (verbose > 2)
-			rprintf(FINFO, "calling match_sums %s\n", fname);
+			rprintf(FINFO, "calling match_sums %s%s%s\n", path,slash,fname);
 
 		if (log_before_transfer)
 			log_item(FCLIENT, file, &initial_stats, iflags, NULL);
 		else if (!am_server && verbose && do_progress)
-			rprintf(FCLIENT, "%s\n", fname2);
+			rprintf(FCLIENT, "%s\n", fname);
 
 		set_compression(fname);
 
@@ -360,7 +350,7 @@ void send_files(struct file_list *flist, int f_out, int f_in)
 			j = unmap_file(mbuf);
 			if (j) {
 				io_error |= IOERR_GENERAL;
-				rsyserr(FERROR, j,
+				rsyserr(FERROR_XFER, j,
 					"read errors mapping %s",
 					full_fname(fname));
 			}
@@ -370,17 +360,18 @@ void send_files(struct file_list *flist, int f_out, int f_in)
 		free_sums(s);
 
 		if (verbose > 2)
-			rprintf(FINFO, "sender finished %s\n", fname);
+			rprintf(FINFO, "sender finished %s%s%s\n", path,slash,fname);
 
 		/* Flag that we actually sent this entry. */
-		file->flags |= FLAG_SENT;
+		file->flags |= FLAG_FILE_SENT;
 	}
-	make_backups = save_make_backups;
+	if (make_backups < 0)
+		make_backups = -make_backups;
 
 	if (verbose > 2)
 		rprintf(FINFO, "send files finished\n");
 
 	match_report();
 
-	write_int(f_out, -1);
+	write_ndx(f_out, NDX_DONE);
 }
