@@ -44,9 +44,9 @@ extern int checksum_seed;
 
 #define XATTR_ABBREV(x) ((size_t)((x).name - (x).datum) < (x).datum_len)
 
-#define XSTATE_ABBREV	0
-#define XSTATE_DONE	1
-#define XSTATE_TODO	2
+#define XSTATE_ABBREV	1
+#define XSTATE_DONE	2
+#define XSTATE_TODO	3
 
 #define USER_PREFIX "user."
 #define UPRE_LEN ((int)sizeof USER_PREFIX - 1)
@@ -178,8 +178,9 @@ static char *get_xattr_data(const char *fname, const char *name, size_t *len_ptr
 
 	if (!datum_len && !extra_len)
 		extra_len = 1; /* request non-zero amount of memory */
-	if (datum_len + extra_len < datum_len /* checks for overflow */
-	 || !(ptr = new_array(char, datum_len + extra_len)))
+	if (datum_len + extra_len < datum_len)
+		overflow_exit("get_xattr_data");
+	if (!(ptr = new_array(char, datum_len + extra_len)))
 		out_of_memory("get_xattr_data");
 
 	if (datum_len) {
@@ -287,6 +288,48 @@ int get_xattr(const char *fname, stat_x *sxp)
 	return 0;
 }
 
+int copy_xattrs(const char *source, const char *dest)
+{
+	ssize_t list_len, name_len;
+	size_t datum_len;
+	char *name, *ptr;
+#ifdef HAVE_LINUX_XATTRS
+	int user_only = am_sender ? 0 : !am_root;
+#endif
+
+	/* This puts the name list into the "namebuf" buffer. */
+	if ((list_len = get_xattr_names(source)) < 0)
+		return -1;
+
+	for (name = namebuf; list_len > 0; name += name_len) {
+		name_len = strlen(name) + 1;
+		list_len -= name_len;
+
+#ifdef HAVE_LINUX_XATTRS
+		/* We always ignore the system namespace, and non-root
+		 * ignores everything but the user namespace. */
+		if (user_only ? !HAS_PREFIX(name, USER_PREFIX)
+			      : HAS_PREFIX(name, SYSTEM_PREFIX))
+			continue;
+#endif
+
+		datum_len = 0;
+		if (!(ptr = get_xattr_data(source, name, &datum_len, 0)))
+			return -1;
+		if (sys_lsetxattr(dest, name, ptr, datum_len) < 0) {
+			int save_errno = errno ? errno : EINVAL;
+			rsyserr(FERROR_XFER, errno,
+				"rsync_xal_set: lsetxattr(\"%s\",\"%s\") failed",
+				dest, name);
+			errno = save_errno;
+			return -1;
+		}
+		free(ptr);
+	}
+
+	return 0;
+}
+
 static int find_matching_xattr(item_list *xalp)
 {
 	size_t i, j;
@@ -460,9 +503,11 @@ int xattr_diff(struct file_struct *file, stat_x *sxp, int find_all)
 	return !xattrs_equal;
 }
 
-/* When called by the generator with a NULL fname, this tells the sender
- * which abbreviated xattr values we need.  When called by the sender
- * (with a non-NULL fname), we send all the extra xattr data it needs. */
+/* When called by the generator (with a NULL fname), this tells the sender
+ * all the abbreviated xattr values we need.  When called by the sender
+ * (with a non-NULL fname), we send all the extra xattr data it needs.
+ * The generator may also call with f_out < 0 to just change all the
+ * XSTATE_ABBREV states into XSTATE_DONE. */
 void send_xattr_request(const char *fname, struct file_struct *file, int f_out)
 {
 	item_list *lst = rsync_xal_l.items;
@@ -481,6 +526,7 @@ void send_xattr_request(const char *fname, struct file_struct *file, int f_out)
 				rxa->datum[0] = XSTATE_DONE;
 			continue;
 		case XSTATE_TODO:
+			assert(f_out >= 0);
 			break;
 		default:
 			continue;
@@ -509,7 +555,8 @@ void send_xattr_request(const char *fname, struct file_struct *file, int f_out)
 		}
 	}
 
-	write_byte(f_out, 0); /* end the list */
+	if (f_out >= 0)
+		write_byte(f_out, 0); /* end the list */
 }
 
 /* When called by the sender, read the request from the generator and mark
@@ -543,8 +590,9 @@ int recv_xattr_request(struct file_struct *file, int f_in)
 				who_am_i(), num, f_name(file, NULL));
 			exit_cleanup(RERR_STREAMIO);
 		}
-		if (rxa->datum_len <= MAX_FULL_DATUM || rxa->datum[0] != XSTATE_ABBREV) {
-			rprintf(FERROR, "[%s] internal abbrev error!\n", who_am_i());
+		if (!XATTR_ABBREV(*rxa) || rxa->datum[0] != XSTATE_ABBREV) {
+			rprintf(FERROR, "[%s] internal abbrev error on %s (%s, len=%ld)!\n",
+				who_am_i(), f_name(file, NULL), rxa->name, (long)rxa->datum_len);
 			exit_cleanup(RERR_STREAMIO);
 		}
 
@@ -557,7 +605,7 @@ int recv_xattr_request(struct file_struct *file, int f_in)
 		rxa->datum_len = read_varint(f_in);
 
 		if (rxa->name_len + rxa->datum_len < rxa->name_len)
-			out_of_memory("recv_xattr_request"); /* overflow */
+			overflow_exit("recv_xattr_request");
 		rxa->datum = new_array(char, rxa->datum_len + rxa->name_len);
 		if (!rxa->datum)
 			out_of_memory("recv_xattr_request");
@@ -609,10 +657,9 @@ void receive_xattr(struct file_struct *file, int f)
 		size_t datum_len = read_varint(f);
 		size_t dget_len = datum_len > MAX_FULL_DATUM ? 1 + MAX_DIGEST_LEN : datum_len;
 		size_t extra_len = MIGHT_NEED_RPRE ? RPRE_LEN : 0;
-		if (dget_len + extra_len < dget_len)
-			out_of_memory("receive_xattr"); /* overflow */
-		if (dget_len + extra_len + name_len < dget_len)
-			out_of_memory("receive_xattr"); /* overflow */
+		if ((dget_len + extra_len < dget_len)
+		 || (dget_len + extra_len + name_len < dget_len))
+			overflow_exit("receive_xattr");
 		ptr = new_array(char, dget_len + extra_len + name_len);
 		if (!ptr)
 			out_of_memory("receive_xattr");
